#!/usr/bin/env python

from t2t.remap import parse_otu_map, members_to_rep, remap_taxonomy
from unittest import TestCase, main


class RemapTests(TestCase):
    def test_parse_otu_map(self):
        mapping = ("1\t2\t3\t4\n"
                   "2\t5\t6\n"
                   "3\t7")
        exp = {"2": ["2", "3", "4"],
               "5": ["5", "6"],
               "7": ["7"]}
        obs = parse_otu_map(mapping.splitlines())
        self.assertEqual(obs, exp)

    def test_members_to_rep(self):
        otus = {"2": ["2", "3", "4"],
                "5": ["5", "6"],
                "7": ["7"]}
        exp = {'2': '2', '3': '2', '4': '2', '5': '5', '6': '5',
               '7': '7'}
        obs = members_to_rep(otus)
        self.assertEqual(obs, exp)

    def test_remap_taxonomy(self):
        mapping= {"2": ["2", "3", "4"],
                  "5": ["5", "6"],
                  "7": ["7"]}

        tax = {"3": ["a", "b", "c"],
               "7": ["x", "y", "z"],
               "8": ['foo', 'bar']}

        exp = {"2": ["a", "b", "c"],
               "3": ["a", "b", "c"],
               "4": ["a", "b", "c"],
               "7": ["x", "y", "z"],
               "8": ['foo', 'bar']}

        obs = remap_taxonomy(mapping, tax)
        self.assertEqual(obs, exp)

if __name__ == '__main__':
    main()



#!/usr/bin/env python

from collections import defaultdict
from operator import itemgetter
from numpy import argmin, array, where
from skbio import TreeNode
from t2t.util import unzip
import re

__author__ = "Daniel McDonald"
__copyright__ = "Copyright 2011, The tax2tree project"
__credits__ = ["Daniel McDonald"]
__license__ = "BSD"
__version__ = "1.0"
__maintainer__ = "Daniel McDonald"
__email__ = "mcdonadt@colorado.edu"
__status__ = "Development"


RANK_ORDER = ['d', 'p', 'c', 'o', 'f', 'g', 's']
BAD_NAMES = [
    'environmental sample', 'uncultured', 'UNNAMEABLE', 'unclassified',
    'unidentified', 'cluster', 'isolate', 'environmental samples']

BAD_NAMES_REGEX = re.compile("(%s)" % ')|('.join(map(str.lower, BAD_NAMES)))


def set_rank_order(order):
    """Reset the global RANK_ORDER"""
    global RANK_ORDER
    RANK_ORDER = order


def determine_rank_order(con):
    """Determines dynamically rank order based on first input con string"""
    order = [s.strip()[0] for s in con.split(';')]
    global RANK_ORDER
    RANK_ORDER = order

    return order


def has_badname(name):
    """Boolean, if name contains a badname"""
    return len(BAD_NAMES_REGEX.findall(name)) > 0


def load_consensus_map(lines, append_rank, check_bad=True,
                       check_min_inform=True, assert_nranks=True,
                       verbose=False, check_euk_unc=False):
    """Input is tab delimited mapping from tipname to a consensus string

    tipname is the tipnames in the loaded tree
    consensus string must be len(RANK_ORDER), and ';' delimited

    check_bad : check for bad names
    check_min_inform: check for informative information below domain

    If append_rank is True, rank information will be appended on. For instance,
    the name at the 0-index position of the consensus will be joined with
    RANK_ORDER[0]

    check_euk_unc : check for eukarayota or unclassified, set to none if found
    and true

    Output is a dictionary mapping tipname to consensus strings split into
    a list.
    """
    if verbose:
        print("loading consensus map...")

    mapping = {}
    n_ranks = len(RANK_ORDER)
    for line in lines:
        id_, consensus = line.strip().split('\t')
        id_ = id_.strip()

        names = [n.strip() for n in consensus.split(';')]

        if check_euk_unc and 'Eukaryota' in names[0] or \
                'Unclassified' in names[0]:
            names = [None] * n_ranks

        if assert_nranks:
            if len(names) != n_ranks:
                raise ValueError

        # clean up missing names
        for idx in range(n_ranks):
            if names[idx] == '' or names[idx] == 'None':
                names[idx] = None
            if names[idx] is not None and \
                    '__' in names[idx] and \
                    names[idx].split('__')[1] == '':
                names[idx] = None

        if check_min_inform:
            if names[1] is None:
                names = [None] * n_ranks

        # clean up bad names
        if check_bad:
            for idx, name in enumerate(names):
                if name is not None and has_badname(name.lower()):
                    names[idx] = None

        # append rank if needed
        if append_rank:
            for idx in range(n_ranks):
                if names[idx] is not None:
                    names[idx] = '__'.join([RANK_ORDER[idx], names[idx]])
                else:
                    names[idx] = "%s__" % RANK_ORDER[idx]
        mapping[id_] = names

    return mapping


def load_tree(tree, tipname_map):
    """Returns a PhyloNode tree decorated with helper attrs

    The following attributes and descriptions are decorated onto the tree:

        Consensus
            If the node is a tip, the corresponding taxonomy string is placed
            here otherwise [None] * number_of_ranks is stored

            If the node is internal, then [None] * number_of_ranks is stored

        TipStart
            The left most tip

        TipStop
            The right most tip

    Parameters
    ----------
    tree : str or TreeNode
        A newick string or a TreeNode
    tipname_map : dict
        {id_: [tax, string]}

    Returns
    -------
    TreeNode

    """
    if not isinstance(tree, TreeNode):
        tree = TreeNode.read(tree, convert_underscores=False)

    n_ranks = len(RANK_ORDER)

    missing_tax = [None] * n_ranks

    for idx, tip in enumerate(tree.tips()):
        if tip.name:
            tip.name = tip.name.replace("'", "")

        tip.TipStart = idx
        tip.TipStop = idx
        tip.Consensus = tipname_map.get(tip.name, missing_tax)

    for node in tree.postorder(include_self=True):
        if node.is_tip():
            continue

        node.TipStart = node.children[0].TipStart
        node.TipStop = node.children[-1].TipStop
        node.Consensus = missing_tax

        if node.name is None:
            node.Bootstrap = None
        else:
            try:
                node.Bootstrap = float(node.name)
                node.name = None
            except ValueError:
                node.Bootstrap = None

    return tree


def collect_names_at_ranks_counts(tree):
    """Returns total name counts for a given name at a given rank

    Assumes the Consensus attribute is present on the tips

    Parameters
    ----------
    tree : TreeNode

    Returns
    -------
    dict of dict
        Returns a 2d dict, [RANK][name] -> count

    """
    total_counts = {i: defaultdict(int) for i in range(len(RANK_ORDER))}

    for consensus in (tip.Consensus for tip in tree.tips()):
        for rank, name in enumerate(consensus):
            if name is None:
                continue
            total_counts[rank][name] += 1
    return total_counts


def decorate_name_relative_freqs(tree, total_counts, min_count):
    """Decorates relative frequency information for names on the tree

    Adds on the attribute ConsensusRelFreq which is a 2d dict containing
    the relative frequency of each name at each rank for the subtree that
    descends from a given node.

    Adds on the attribute ValidRelFreq which is a 2d dict containing
    the valid tip frequency of each name at each rank for the subtree that
    descends from a given node.

    Both these attributes will be None on tips.

    Parameters
    ----------
    tree : TreeNode
    total_counts : dict of dict
        The return data from collect_names_at_ranks_counts
    min_count : int
        is the minimum number of tips that must represent a name for that
        frequency to be retained

    """
    tips = list(tree.tips())
    for tip in tips:
        tip.ConsensusRelFreq = None

    n_ranks = len(RANK_ORDER)
    n_ranks_it = range(n_ranks)

    for n in tree.traverse(include_self=True):
        if n.is_tip():
            n.ConsensusRelFreq = None
            n.ValidRelFreq = None
            continue

        counts = {i: defaultdict(int) for i in n_ranks_it}

        # build of counts of the names at the tips per rank
        cons_at_tips = (t.Consensus for t in tips[n.TipStart:n.TipStop + 1])
        for con in cons_at_tips:
            for cur_rank, cur_name in enumerate(con):
                if cur_name is None:
                    continue
                counts[cur_rank][cur_name] += 1

        res_freq = {i: {} for i in n_ranks_it}
        res_valid = {i: {} for i in n_ranks_it}

        # collect frequency information of the names per rank
        for rank, names in counts.items():
            for name, name_counts in counts[rank].items():
                if name_counts < min_count:
                    continue

                relfreq = float(name_counts) / total_counts[rank][name]
                validfreq = float(name_counts) / n.NumTips
                res_freq[rank][name] = relfreq
                res_valid[rank][name] = validfreq

        n.ConsensusRelFreq = res_freq
        n.ValidRelFreq = res_valid


def decorate_name_counts(tree):
    """Decorates count information for names on the tree

    Adds on the attribute TaxaCount which is a 2d dict containing
    the count of each name at each rank for the subtree that
    descends from a given node.

    Parameters
    ----------
    tree : TreeNode
    total_counts : dict of dict
        The return data from collect_names_at_ranks_counts
    """

    tips = list(tree.tips())
    n_ranks = len(RANK_ORDER)
    n_ranks_it = range(n_ranks)

    for n in tree.traverse(include_self=True):
        counts = {i: defaultdict(int) for i in n_ranks_it}

        # build of counts of the names at the tips per rank
        cons_at_tips = (t.Consensus for t in tips[n.TipStart:n.TipStop + 1])
        for con in cons_at_tips:
            for cur_rank, cur_name in enumerate(con):
                if cur_name is None:
                    continue
                counts[cur_rank][cur_name] += 1

        n.TaxaCount = counts


def set_ranksafe(tree):
    """Determines what ranks are safe for a given node

    RankSafe is a len(RANK_ORDER) boolean list. True means at that rank, there
    is only a single name with > 50% relative abundance

    Parameters
    ----------
    tree : TreeNode

    """
    ranksafe = [False] * len(RANK_ORDER)
    for node in tree.traverse(include_self=True):
        node.RankSafe = ranksafe[:]

        if node.is_tip():
            continue

        for rank, names in node.ConsensusRelFreq.items():
            # this is strict
            if sum(x >= 0.5 for x in names.values()) == 1:
                node.RankSafe[rank] = True


def decorate_ntips(tree):
    """Cache the number of informative tips on the tree.

    This method will set NumTips as the number of informative tips that descend
    from a given node. If the node is a tip, and it is informative, it will
    have a count of 1. Informative is based on the presence of taxonomy
    information at a tip.

    Parameters
    ----------
    tree : TreeNode

    """
    n_ranks = len(RANK_ORDER)
    missing = [None] * n_ranks

    for node in tree.postorder(include_self=True):
        if node.is_tip():
            node.NumTips = node.Consensus != missing
        else:
            node.NumTips = sum(c.NumTips for c in node.children)


def decorate_ntips_rank(tree):
    """Cache the number of informative tips for each rank for each node.

    This method will set NumTipsRank as the number of informative tips that
    descend from a given node for each rank. Informative is based on the
    presence of taxonomy information at a tip for a give rank.

    Parameters
    ----------
    tree : TreeNode

    """
    n_ranks = len(RANK_ORDER)

    for node in tree.postorder(include_self=True):
        counts = defaultdict(int)
        for r in range(n_ranks):
            if node.is_tip():
                counts[r] = node.Consensus[r] is not None
            else:
                counts[r] = sum(c.NumTipsRank[r] for c in node.children)

        node.NumTipsRank = counts


def pick_names(tree):
    """Does an initial decoration of names on the tree

    The best name by relative frequency is placed, from kingdom -> species,
    up until the first uninformative name following an informative name is
    placed

    """
    names_prealloc = [None] * len(RANK_ORDER)

    for node in tree.non_tips(include_self=True):
        names = names_prealloc[:]
        count = 0

        # set names at ranksafe nodes, stop if we've set a name and descendent
        # rank names are not safe.
        for rank, is_safe in enumerate(node.RankSafe):
            if is_safe:
                # place best name
                count += 1
                relfreq = node.ConsensusRelFreq[rank].items()
                names[rank] = sorted(relfreq, key=itemgetter(1))[-1][0]
            else:
                # if we've had one or more useless rank, set remaining to None
                if count >= 1:
                    break

        node.RankNames = names


def fmeasure(precision, recall):
    """Returns the fmeasure (or F1-score)

    http://en.wikipedia.org/wiki/F1_score
    """
    return 2.0 * ((precision * recall) / (precision + recall))


def fpoint5measure(precision, recall):
    """Returns the f0.5measure (or F0.5-score)"""
    beta = 0.5
    betasqrd = beta ** 2

    tmp = (precision * recall) / ((betasqrd * precision) + recall)
    return (1 + betasqrd) * tmp


def f2measure(precision, recall):
    """Returns the f2measure (or F2-score)"""
    return (1.0 + (2 ** 2)) * ((precision * recall) /
                               ((2 ** 2 * precision) + recall))


def min_tips(nodes):
    """For a list of nodes, return the node with the fewest tips

    implemented naively...
    """
    scores = []
    for n in nodes:
        if n is None:
            scores.append(99999999999)
        else:
            scores.append(len(list(n.tips())))
    return nodes[argmin(scores)]


def name_node_score_fold(tree, score_f=fmeasure, tiebreak_f=min_tips,
                         verbose=False):
    """Compute name scores for internal nodes, pick the 'best'

    For this method, we traverse the tree once building up a dict of scores
    for names and nodes, we can then pick the 'best' node out of the dict
    to avoid horrible lookups in the tree
    """

    if verbose:
        print("Starting name_node_score_fold...")

    name_node_score = {i: {} for i in range(len(RANK_ORDER))}
    n_ranks = len(RANK_ORDER)

    for node in tree.non_tips(include_self=True):
        node.RankNameScores = [None] * n_ranks

        for rank, name in enumerate(node.RankNames):
            if name is None:
                continue

            # precision in this case is the percent of informative tips that
            # descend that are of the name relative to the number of
            # informative tips that descend
            precision = node.ValidRelFreq[rank][name]

            # recall in this case is the percent of informative tips that
            # descent that are of the name relative to the total number of
            # tips in the tree with name
            recall = node.ConsensusRelFreq[rank][name]

            # calculate score and save it for the corrisponding rank position
            # so that these values can be examined later in other contexts
            score = score_f(precision, recall)
            node.RankNameScores[rank] = score

            if name not in name_node_score[rank]:
                name_node_score[rank][name] = []
            name_node_score[rank][name].append((node, score))

    # run through the built up dict and pick the best node for a name
    used_scores = {}
    for rank, names in name_node_score.items():
        used_scores[rank] = []

        for name, node_scores in names.items():
            node_scores_sorted = sorted(node_scores, key=itemgetter(1))[::-1]
            nodes, scores = unzip(node_scores_sorted)
            scores = array(scores)

            used_scores[rank].append((name, scores[0]))

            # if there is a tie in scores...
            if sum(scores == scores[0]) > 1:
                # ugly hack to get around weird shape mismatch
                indices = where(scores == scores[0], range(len(nodes)), None)
                tie_nodes = []
                for i in indices:
                    if i is not None:
                        tie_nodes.append(nodes[i])
                    else:
                        tie_nodes.append(None)
                node_to_keep = tiebreak_f(tie_nodes)
                for node, score in node_scores_sorted:
                    if node == node_to_keep:
                        continue
                    else:
                        node.RankNames[rank] = None
            else:
                for node, score in node_scores_sorted[1:]:
                    node.RankNames[rank] = None

    return used_scores


def score_tree(tree, verbose=False):
    """Scores the tree based on RankNameScores and tip coverage

    if the node has a name in RankNames, multiply score in RankNameScores @ idx
    by the number of tips that are "valid".. Sum these scores, divide sum by
    total number of tips covered (a tip can be counted multiple times)
    """
    total_score = 0.0
    tip_count = 0

    if verbose:
        print("Scoring tree...")

    for n in tree.non_tips(include_self=True):
        for idx, name in enumerate(n.RankNames):
            if name is None:
                continue
            score = n.RankNameScores[idx]
            total_score += score * n.NumTips
            tip_count += n.NumTips
    return total_score / tip_count


def set_preliminary_name_and_rank(tree):
    """Sets names and rank at a node

    This method is destructive: will destroy the Name attribute on tree
    """

    n_ranks = len(RANK_ORDER)
    empty_ranknames = [None] * n_ranks

    for node in tree.non_tips(include_self=True):
        node.name = None
        node.Rank = None

        if empty_ranknames == node.RankNames:
            continue

        # take the "deepest" name
        for idx, name in enumerate(node.RankNames[::-1]):
            if name is None:
                continue
            node.name = name
            node.Rank = n_ranks - (idx + 1)  # adjust for 0-based index
            break


def make_consensus_tree(cons_split, check_for_rank=True, tips=None):
    """Returns a mapping by rank for names to their parent names and counts"""

    god_node = TreeNode(name=None)
    god_node.Rank = None

    base = list(cons_split)[0]
    #  SMJ base = next(iter(cons_split))
    cur_node = god_node

    # create a base path in the tree
    for rank, name in enumerate(base):
        new_node = TreeNode(name=name)
        new_node.Rank = rank
        cur_node.append(new_node)
        cur_node = new_node

    # setup the initial childlookup structure so taht we don't have to
    # always iterate over .children
    for n in god_node.traverse(include_self=True):
        if n.is_tip():
            n.ChildLookup = {}
            continue
        n.ChildLookup = {n.children[0].name: n.children[0]}

    # for every consensus string, start at the "god" node
    for idx, con in enumerate(cons_split):
        cur_node = god_node

        # for each name, see if we've seen it, if not, add that puppy on
        for rank, name in enumerate(con):
            if name in cur_node.ChildLookup:
                cur_node = cur_node.ChildLookup[name]
            else:
                new_node = TreeNode(name=name)
                new_node.Rank = rank
                new_node.ChildLookup = {}
                cur_node.append(new_node)
                cur_node.ChildLookup[name] = new_node
                cur_node = new_node

        if tips is not None:
            cur_node.append(TreeNode(name=tips[idx]))

    # build an assist lookup dict
    lookup = {}
    for node in god_node.traverse():
        if node.name is None:
            continue
        if check_for_rank and '__' in node.name and \
                node.name.split('__')[1] == '':
            continue
        lookup[node.name] = node

    return god_node, lookup


def get_nearest_named_ancestor(node):
    """Returns node of nearest .name'd ancestor

    returns None if there does not exist a named ancestor
    """
    curr = node.parent
    while curr is not None and curr.name is None:
        curr = curr.parent
    return curr


def backfill_names_gap(tree, consensus_lookup, verbose=False):
    """Fill in missing names

    use the consensus tree mapped by nodes_lookup to attempt to fill in missing
    consensus names. For instance, if we know, by the the consensus tree,
    that foo has the parent bar, then if we see the name foo in the real tree
    and it does not have a parent named bar, we can add that name on.

    We set the attribute BackFillNames here as we want to attempt to collapse
    names later if we sanely and easily can
    """
    for node in tree.non_tips(include_self=True):

        if node.name is not None:
            node.BackFillNames = [node.name]
        else:
            node.BackFillNames = []
            continue

        if node.parent is None:
            continue
        if node.Rank is None:
            continue
        if node.is_tip():
            continue
        # if node is kingdom, more on
        if node.Rank == 0:
            continue

        # find nearest ranked parent
        named_ancestor = get_nearest_named_ancestor(node)

        if named_ancestor is None:
            if verbose:
                print("Unable to find a named parent for %s" % (node.name))
            continue

        levels = node.Rank - named_ancestor.Rank

        # this is indicative of a problem with the consensus strings
        if levels < 0:
            print(node.name, node.Rank, named_ancestor.Rank)
            print('\t', node.RankSafe)
            print('\t', node.RankNames)
            print('\t', named_ancestor.RankSafe)
            print('\t', named_ancestor.RankNames)
            node.BackFillNames = []
            continue
        elif levels == 1:
            continue

        # walk consensus tree for missing names
        names = walk_consensus_tree(consensus_lookup, node.name, levels)
        node.BackFillNames = names


def walk_consensus_tree(lookup, name, levels, reverse=True, verbose=False):
    """Walk up the consensus tree for n levels, return names

    if reverse is True, names are [::-1]
    """
    node = lookup[name]
    names = [name]
    curr = node.parent

    for _ in range(1, levels):
        if curr.Rank is None:
            # at root...
            break

        if curr is None:
            if verbose:
                print("Possible malformed consensus tree! See node %s" % name)
            continue
        if curr.name is None:
            if verbose:
                print("Gap in consensus tree! See node %s" % name)
            names.append('%s__' % RANK_ORDER[curr.Rank])
        else:
            names.append(curr.name)
        curr = curr.parent

    if reverse:
        names = names[::-1]

    return names


def commonname_promotion(tree):
    """Promote names if possible from BackFillNames"""
    for node in tree.preorder(include_self=True):
        queue = [c for c in node.children[:] if not c.is_tip()]
        backfill_nodes = []
        push_name = True

        # collect nearest descendants backfilled names
        while queue:
            curr = queue.pop()
            if len(curr.BackFillNames) == 0:
                queue.extend([c for c in curr.children[:] if not c.is_tip()])
            elif len(curr.BackFillNames) == 1:
                push_name = False
                break
            else:
                backfill_nodes.append(curr)

        # see if there is 100% overlap in a name at a given rank, if so
        # we can and should put it on the deeper node
        while push_name:
            cur_name = [n.BackFillNames[0] for n in backfill_nodes]
            if len(set(cur_name)) == 1:
                node.BackFillNames.append(cur_name[0])

                for n in backfill_nodes:
                    n.BackFillNames.pop(0)
                    if len(n.BackFillNames) <= 1:
                        push_name = False
            else:
                push_name = False

    # set the .name attribute on the tree based on .BackFillNames
    for node in tree.preorder(include_self=True):
        if node.is_tip():
            continue

        if node.BackFillNames:
            unique = []
            for name in node.BackFillNames:
                unique.append(TaxaName.getTaxaName(name))
            node.BackFillNames = unique
        if len(node.BackFillNames) > 1:
            node.name = '; '.join(node.BackFillNames)
        elif len(node.BackFillNames) == 1:
            node.name = TaxaName.getTaxaName(node.BackFillNames[0])
        else:
            node.name = None


class TaxaName(object):

    """Support object to provide unique taxa names"""
    _names = {}

    def __init__(self):
        raise NotImplementedError("Object not meant for instantiation")

    @classmethod
    def getTaxaName(cls, request):
        return request


def make_names_unique(tree, append_suffix=True, suffix_glue_char='_',
                      verbose=False):
    """Appends on a unique number if multiple of the same names exist

    ordered by number of tips, ie, _1 has more tips that _2

    expects .BackFillNames to be set
    """
    if verbose:
        print("Assigning unique tags to duplicate names...")

    # build up a dict of the names and how many tips descend
    name_lookup = {}
    for node in tree.non_tips(include_self=True):
        if node.name is None:
            continue
        else:
            for idx, name in enumerate(node.BackFillNames):
                if name not in name_lookup:
                    name_lookup[name] = []
                name_info = ((node.TipStop - node.TipStart), idx, node)
                name_lookup[name].append(name_info)

    # assign unique numbers based on the number of tips that descend
    for name, scores_and_nodes in name_lookup.items():
        sorted_scores = sorted(scores_and_nodes, key=lambda x: x[0])[::-1]
        for count, (score, idx, node) in enumerate(sorted_scores):
            # only assign a number of we have more than 1
            if count > 0:
                if node.BackFillNames[idx].split('__')[1] != '':
                    if append_suffix:
                        unique_name = suffix_glue_char.join(
                            [node.BackFillNames[idx], str(count)])
                        node.BackFillNames[idx] = unique_name

    # should probably be refactored, but assign .name based on .BackFillNames
    for node in tree.non_tips(include_self=True):
        if len(node.BackFillNames) == 0:
            node.name = None
        elif len(node.BackFillNames) == 1:
            node.name = node.BackFillNames[0]
        else:
            node.name = '; '.join(node.BackFillNames)


def pull_consensus_strings(tree, verbose=False, append_prefix=True):
    """Pulls consensus strings off of tree

    assumes .name is set
    """
    if verbose:
        print("Pulling consensus strings...")

    constrings = []
    rank_order_rev = {r: i for i, r in enumerate(RANK_ORDER)}
    # start at the tip and travel up
    for tip in tree.tips():
        if append_prefix:
            consensus_string = ['%s__' % r for r in RANK_ORDER]
        else:
            consensus_string = ['' for r in RANK_ORDER]

        tipid = tip.name
        n = tip.parent

        # walk up the tree filling in the consensus string
        while n.parent:
            if n.name:
                if ';' in n.name:
                    names = [r.strip() for r in n.name.split(';')]
                    for name in names:
                        rank_idx = rank_order_rev[name[0]]
                        consensus_string[rank_idx] = name
                else:
                    rank_idx = rank_order_rev[n.name[0]]
                    consensus_string[rank_idx] = n.name
            n = n.parent

        # if there is a name at the root we need to make sure we grab it
        if n.name:
            if ';' in n.name:
                names = [r.strip() for r in n.name.split(';')]
                for name in names:
                    rank_idx = rank_order_rev[name[0]]
                    consensus_string[rank_idx] = name
            else:
                rank_idx = rank_order_rev[n.name[0]]
                consensus_string[rank_idx] = n.name

        # join strings with tip id
        constrings.append('\t'.join([tipid, '; '.join(consensus_string)]))
    return constrings


def save_bootstraps(tree, verbose=False):
    """Retains .Bootstrap if set in .name"""
    if verbose:
        print("Attempting to retain bootstrap values")
    for n in tree.non_tips(include_self=True):
        if n.Bootstrap is not None:
            if n.name is None:
                n.name = str(n.Bootstrap)
            else:
                if ';' in n.name:
                    n.name = ':'.join([str(n.Bootstrap), "%s" % n.name])
                else:
                    n.name = ':'.join([str(n.Bootstrap), n.name])


def getpath(foo):
    while foo.parent:
        print(foo.name)
        if hasattr(foo, 'RankNames'):
            print(foo.RankNames)
        foo = foo.parent
    print(foo.name)
    if hasattr(foo, 'RankNames'):
        print(foo.RankNames)


def is_float(s):
    """Returns True if the value in string s is a float"""
    if s is None:
        return False

    if '.' not in s:
        return False

    try:
        float(s)
        return True
    except ValueError:
        return False


def validate_all_paths(tree):
    """Walk each path in the tree and make sure there aren't any conflicts"""
    # helper getpath method
    def getpath_f(n):
        path = []
        n = n.parent
        while n.parent:
            if n.name is not None:
                if ':' in n.name:
                    names = n.name.split(':', 1)
                    if not is_float(names[1]):
                        path.append(names[1])
                else:
                    if not is_float(n.name):
                        path.append(n.name)
            n = n.parent
        if n.name is not None:
            path.append(n.name)

        clean_path = []
        for p in path:
            if ';' in p:
                clean_path.extend(p.split(';')[::-1])
            else:
                clean_path.append(p)
        return clean_path

    rank_order_rev = {r: i for i, r in enumerate(RANK_ORDER)}

    bad_tips = []

    for tip in tree.tips():
        tip_path = getpath_f(tip)

        # look up the ranks on the names
        vals = [rank_order_rev[p[0]] for p in tip_path]

        # sorting shouldn't change order, if it does, ranks are out of order
        if sorted(vals)[::-1] != vals:
            bad_tips.append(tip)

        # casting to a set shouldn't change the number of ranks present,
        # if it does we have a duplicate rank
        elif len(vals) != len(set(vals)):
            bad_tips.append(tip)

    return bad_tips



#!/usr/bin/env python

__author__ = "Daniel McDonald"
__copyright__ = "Copyright 2011, The tax2tree project"
__credits__ = ["Daniel McDonald"]
__license__ = "BSD"
__version__ = "1.0"
__maintainer__ = "Daniel McDonald"
__email__ = "mcdonadt@colorado.edu"
__status__ = "Development"

__all__ = ['nlevel', 'reroot', 'consensus']



#!/usr/bin/env python


def parse_otu_map(lines):
    """Returns {rep: [members]}, members include rep"""
    res = {}
    for line in lines:
        fields = line.strip().split('\t')
        rep = fields[1]
        members = fields[1:]
        res[rep] = members
    return res


def members_to_rep(otus):
    """Provides a lookup for a cluster member to its rep"""
    res = {}
    for rep, members in otus.items():
        for member in members:
            res[member] = rep
    return res


def remap_taxonomy(mapping, taxa):
    """Remaps the taxonomy over the OTU clusters"""
    res = {}
    reps = members_to_rep(mapping)

    for tax_id, tax_str in taxa.items():
        rep = reps.get(tax_id, None)

        if rep is None:
            res[tax_id] = tax_str
            continue

        for m in mapping[rep]:
            if m in res:
                print("%s was: %s\nnow is %s" % (m, mapping[m], tax_str))
            res[m] = tax_str
    return res



#!/usr/bin/env python

__author__ = "Daniel McDonald"
__copyright__ = "Copyright 2011, The tax2tree project"
__credits__ = ["Daniel McDonald"]
__license__ = "BSD"
__version__ = "1.0"
__maintainer__ = "Daniel McDonald"
__email__ = "mcdonadt@colorado.edu"
__status__ = "Development"


def reroot(tree, tipnames, tmp_nodename="TEMPORARY_ROOT_NODE_NAME"):
    """Returns a tree rerooted based on tipnames"""
    node = tree.lowest_common_ancestor(tipnames)

    # make a new node that sits inbetween LCA and parent
    parent = node.parent
    parent.remove(node)
    node.parent = None
    new_node = parent.__class__()
    new_node.name = tmp_nodename

    if hasattr(new_node, 'length') and new_node.length:
        new_node.length = node.length / 2.0
        node.length = node.length / 2.0
    else:
        new_node.length = 0.0

    # add node back to tree and reconnect LCA
    parent.append(new_node)
    new_node.append(node)

    # root at the new node, unset its temporary name
    new_tree = tree.root_at(tmp_nodename)
    new_root = new_tree.find(tmp_nodename)
    new_tree.name = None
    new_root.name = None

    # collapse single descendents if they exist
    new_tree.prune()

    return new_tree


def unzip(items):
    """The inverse of zip

    Parameters
    ----------
    items : a nested iteratable

    Returns
    -------
    list
        The unzipped items

    Examples
    --------
    >>> from t2t.util import unzip
    >>> unzip([[1, 2], ['a', 'b']])
    [[1, 'a'], [2, 'b']]

    """
    if items:
        return [list(i) for i in zip(*items)]
    else:
        return []



#!/usr/bin/env python

from unittest import TestCase, main

from skbio import TreeNode

import sys

if sys.version_info[0] == 2:
    from StringIO import StringIO
else:
    from io import StringIO

from t2t.validate import (check_parse, check_n_levels, check_gap,
                          check_prefixes, ParseError, cache_tipnames,
                          get_polyphyletic, find_gap)


class VerifyTaxonomy(TestCase):
    def setUp(self):
        pass

    def test_check_parse(self):
        """returns valid parsed or raises"""
        exp = ("1", ("k__a", "p__b", "c__c", "o__d", "f__e", "g__f", "s__g"))
        obs = check_parse(good_string)
        self.assertEqual(obs, exp)

        obs = check_parse(good_string_2)
        self.assertEqual(obs, exp)

        self.assertRaises(ParseError, check_parse, bad_string)

    def test_check_n_levels(self):
        """requires N levels, or unclassified"""
        id_, parsed = check_parse(good_string)
        self.assertTrue(check_n_levels(parsed, 7))

        id_, parsed = check_parse(good_string_2)
        self.assertTrue(check_n_levels(parsed, 7))

        self.assertFalse(check_n_levels(parsed, 8))

        id_, parsed = check_parse(good_unclassified)
        self.assertTrue(check_n_levels(parsed, 7))

        id_, parsed = check_parse(bad_unclassified1)
        self.assertFalse(check_n_levels(parsed, 7))

        id_, parsed = check_parse(bad_unclassified2)
        self.assertFalse(check_n_levels(parsed, 7))

        id_, parsed = check_parse(bad_unclassified3)
        self.assertFalse(check_n_levels(parsed, 7))

        id_, parsed = check_parse(bad_unclassified4)
        self.assertFalse(check_n_levels(parsed, 7))

    def test_check_gap(self):
        """check if a gap exists in a string"""
        id_, parsed = check_parse(good_string)
        self.assertTrue(check_gap(parsed))

        id_, parsed = check_parse(good_string_2)
        self.assertTrue(check_gap(parsed))

        id_, parsed = check_parse(good_trailing)
        self.assertTrue(check_gap(parsed))

        id_, parsed = check_parse(gap)
        self.assertFalse(check_gap(parsed))

    def test_find_gap(self):
        good_string_idx = -1
        gap_idx = 2
        trailing_idx = -1

        id_, parsed = check_parse(good_string)
        self.assertEqual(find_gap(parsed), good_string_idx)

        id_, parsed = check_parse(good_trailing)
        self.assertEqual(find_gap(parsed), trailing_idx)

        id_, parsed = check_parse(gap)
        self.assertEqual(find_gap(parsed), gap_idx)

    def test_check_prefixes(self):
        """Verify the expected prefixes are present"""
        prefixes = ['k', 'p', 'c', 'o', 'f', 'g', 's']
        id_, parsed = check_parse(good_string)
        self.assertTrue(check_prefixes(parsed, prefixes))

        id_, parsed = check_parse(good_string_2)
        self.assertTrue(check_prefixes(parsed, prefixes))

        id_, parsed = check_parse(bad_prefix)
        self.assertFalse(check_prefixes(parsed, prefixes))

        id_, parsed = check_parse(no_prefix)
        self.assertFalse(check_prefixes(parsed, prefixes))

    def test_cache_tipnames(self):
        """caches tipnames"""
        t = TreeNode.read(StringIO(u"((a,b)c,(d,e)f)g;"))
        cache_tipnames(t)

        self.assertEqual(t.tip_names, ['a', 'b', 'd', 'e'])
        self.assertEqual(t.children[0].tip_names, ['a', 'b'])
        self.assertEqual(t.children[1].tip_names, ['d', 'e'])

    def test_get_polyphyletic(self):
        """get polyphyletic groups"""
        cons = {'a': ['K', 'X1', 'X'],
                'b': ['K', 'X1', 'X'],
                'd': ['K', 'X1', 'Y'],
                'e': ['K', 'X1', 'Y'],
                'g': ['K', 'X2', 'Z'],
                'h': ['K', 'X2', 'Z'],
                'i': ['K', 'X2', 'X'],
                'j': ['K', 'X2', 'X']}

        obs_poly = get_polyphyletic(cons)

        self.assertEqual(len(obs_poly), 6)
        self.assertEqual(sorted(obs_poly[('X', 2)].keys()), ['X1', 'X2'])
        self.assertEqual(sorted(obs_poly[('Y', 2)].keys()), ['X1'])
        self.assertEqual(sorted(obs_poly[('Z', 2)].keys()), ['X2'])
        self.assertEqual(sorted(obs_poly[('X1', 1)].keys()), ['K'])
        self.assertEqual(sorted(obs_poly[('X2', 1)].keys()), ['K'])
        self.assertEqual(sorted(obs_poly[('K', 0)].keys()), [None])

good_string = "1\tk__a; p__b; c__c; o__d; f__e; g__f; s__g"
good_string_2 = "1\tk__a;p__b;c__c;o__d;f__e;g__f;s__g"

# space instead of tab
bad_string = "2 k__a; p__b; c__c; o__d; f__e; g__f; s__g"

gap = "20	k__a; p__b; c__; o__d; f__e; g__f; s__g"
good_trailing = "30	k__a; p__b; c__c; o__d; f__e; g__; s__"
good_unclassified = "30	k__; p__; c__; o__; f__; g__; s__"
bad_unclassified1 = "40	Unclassified"
bad_unclassified2 = "50	k__x; unclassified"
bad_unclassified3 = "60	k__x; "
bad_unclassified4 = "70	k__x"
bad_nlevels = "80	k__a; p__b; c__c"
bad_prefix = "1	k__a; p__b; c__c; q__d; f__e; g__f; s__g"
no_prefix = "1	k__a; b; c; d; e; f; g"


if __name__ == '__main__':
    main()



#!/usr/bin/env python

from collections import defaultdict

from numpy import mean

__author__ = "Donovan Park"
__copyright__ = "Copyright 2014, The tax2tree project"
__credits__ = ["Donovan Park"]
__license__ = "BSD"
__version__ = "1.0"
__maintainer__ = "Donovan Park"
__email__ = "donovan.parks@gmail.com"
__status__ = "Development"


class Consistency(object):
    """Calculates the consistency of taxonomic groups within a reference tree.

    Consistency is a measure between 0 and 1 indicating how close a taxonomic
    group is to being monophyletic.
    """
    def __init__(self, taxa_counts, n_ranks):
        """Initialize class.

        Parameters
        ----------
        taxa_counts : dict of dict
            Maps taxonomic rank and taxon names to counts, [RANK][name] ->
            count
        n_ranks : Init
            Indicates number of ranks in taxonomic hierarchy
        """
        self.taxa_counts = taxa_counts
        self.n_ranks = n_ranks

    def calculate(self, tree, rooted):
        """Return taxonomic consistency of each taxa in tree.

        Consistency is at every node and the highest consistency
        reported for each taxa.

        Parameters
        ----------
        tree : TreeNode
        rooted : Boolean
            Indicates if tree should be treated as rooted
        """

        # determine total number of informative tips in tree for each rank
        total_informative_tips = defaultdict(int)
        for n in tree.tips():
            for rank in range(self.n_ranks):
                total_informative_tips[rank] += n.NumTipsRank[rank]

        # determine highest consistency node for each taxa
        consistency_index = {i: defaultdict(int) for i in range(self.n_ranks)}
        for n in tree.traverse(include_self=True):
            for rank in range(self.n_ranks):
                for name, total_taxa_cnt in self.taxa_counts[rank].items():
                    node_taxa_count = n.TaxaCount[rank].get(name, 0)
                    incongruent_taxa = n.NumTipsRank[rank] - node_taxa_count

                    c = float(node_taxa_count) / (total_taxa_cnt +
                                                  incongruent_taxa)
                    if c > consistency_index[rank][name]:
                        consistency_index[rank][name] = c

                    if not rooted:
                        # consider consistency of taxa in other subtree since
                        # the tree is unrooted
                        node_taxa_count = total_taxa_cnt - node_taxa_count
                        incongruent_taxa = total_informative_tips[rank] - \
                            n.NumTipsRank[rank] - \
                            node_taxa_count
                        c = float(node_taxa_count) / (total_taxa_cnt +
                                                      incongruent_taxa)
                        if c > consistency_index[rank][name]:
                            consistency_index[rank][name] = c

        return consistency_index

    def write_taxon_consistency(self, output_file, consistency_index):
        """Write consistency of each taxon to file.

        Parameters
        ----------
        output_file : TreeNode
        consistency_index : dict of dict
          Taxonomic consistency returned by Consistency.calculate()
        """

        fout = open(output_file, 'w')
        fout.write('Taxon\tCount\tConsistency\n')
        for rank in range(self.n_ranks):
            for name, consistency in consistency_index[rank].items():
                fout.write('%s\t%d\t%.3f\n' % (name,
                                               self.taxa_counts[rank][name],
                                               consistency))

        fout.close()

    def write_rank_consistency(self, output_file, consistency_index, min_taxa,
                               rank_order):
        """Write average consistency of each rank to file.

        Parameters
        ----------
        output_file : TreeNode
        consistency_index : dict of dict
          Taxonomic consistency returned by Consistency.calculate()
        min_taxa: minimum taxa in group for inclusion in calculated average
        """

        fout = open(output_file, 'w')
        fout.write('Rank #\tRank prefix\t# taxon\tAverage consistency\n')
        for rank in range(self.n_ranks):
            val = []
            for name, consistency in consistency_index[rank].items():
                if self.taxa_counts[rank][name] >= min_taxa:
                    val.append(consistency)

            if len(val) > 0:
                fout.write('%s\t%s\t%d\t%.3f\n' % (rank, rank_order[rank],
                                                   len(val), mean(val)))
            else:
                fout.write('%s\t%s\t%d\t%s\n' % (rank, rank_order[rank],
                                                 len(val), 'NA'))

        fout.close()



#!/usr/bin/env python

from collections import defaultdict
from operator import add
from functools import reduce

from t2t.nlevel import (determine_rank_order,
                        make_consensus_tree,
                        load_consensus_map)
from t2t.util import unzip


class ParseError(Exception):
    pass


def check_parse(line):
    """Make sure the line can parse into (id_, [some,thing])"""
    try:
        id_, initial = line.strip().split('\t')
    except ValueError:
        raise ParseError("Unable to split in tab")

    parsed = tuple([n.strip() for n in initial.split(";")])

    if not len(parsed):
        raise ParseError("Line appears to not have a taxonomy")

    return (id_, parsed)


def check_n_levels(parsed, n):
    """Make sure there are n levels in parsed"""
    if len(parsed) == n:
        return True
    else:
        return False


def find_gap(parsed):
    """Find a gap else -1"""
    reduced = [p.split('__')[-1] for p in parsed]

    end_found = False
    end_idx = -1

    for idx, n in enumerate(reduced):
        if n and end_found:
            break

        if not n:
            end_found = True
            end_idx = idx

    return end_idx if end_idx != len(parsed) - 1 else -1


def check_gap(parsed):
    return find_gap(parsed) == -1


def check_prefixes(parsed, expected_prefixes):
    """Make sure each rank has the expected prefix"""
    for name, prefix in zip(parsed, expected_prefixes):
        try:
            obs, level_name = name.split('__', 1)
        except ValueError:
            return False

        if obs != prefix:
            return False

    return True


def cache_tipnames(t):
    """cache tipnames on the internal nodes"""
    for n in t.postorder(include_self=True):
        if n.is_tip():
            n.tip_names = [n.name]
        else:
            n.tip_names = reduce(add, [c.tip_names for c in n.children])


def get_polyphyletic(cons):
    """get polyphyletic groups and a representative tip"""
    tips, taxonstrings = unzip(cons.items())
    tree, lookup = make_consensus_tree(taxonstrings, False, tips=tips)
    cache_tipnames(tree)

    names = {}
    for n in tree.non_tips():
        if n.name is None:
            continue
        if (n.name, n.Rank) not in names:
            names[(n.name, n.Rank)] = {}
        if n.parent is not None:
            names[(n.name, n.Rank)][n.parent.name] = n.tip_names[0]

    return names


def hierarchy_errors(tax_lines):
    """Get errors in the taxonomy hierarchy"""
    conmap = load_consensus_map(tax_lines, False)
    names = get_polyphyletic(conmap)
    errors = []

    for (name, rank), parents in names.iteritems():
        if len(parents) > 1:
            err = {'Taxon': name, 'Rank': rank, 'Parents': parents}
            errors.append(err)

    return errors


def flat_errors(tax_lines):
    """Flat file errors"""
    inc_prefix = 'Incorrect prefixes'
    inc_nlevel = 'Incorrect number of levels'
    inc_gap = 'Gaps in taxonomy'

    seed_con = tax_lines[0].strip().split('\t')[1]
    rank_order = determine_rank_order(seed_con)

    nlevels = len(rank_order)
    errors = defaultdict(list)
    errors_seen = defaultdict(set)

    for line in tax_lines:
        id_, parsed = check_parse(line)

        if not check_prefixes(parsed, rank_order):
            if parsed not in errors_seen[inc_prefix]:
                errors_seen[inc_prefix].add(parsed)
                errors[inc_prefix].append(id_)

        if not check_n_levels(parsed, nlevels):
            if parsed not in errors_seen[inc_nlevel]:
                errors_seen[inc_nlevel].add(parsed)
                errors[inc_nlevel].append(id_)

        if not check_gap(parsed):
            gap_idx = find_gap(parsed)
            taxon_following_gap = gap_idx + 1

            # another +1 as the slice is exclusive
            if parsed[:taxon_following_gap + 1] not in errors_seen[inc_gap]:
                errors_seen[inc_gap].add(parsed[:taxon_following_gap + 1])
                errors['Gaps in taxonomy'].append(id_)

    return errors



#!/usr/bin/env python

"""TODO: place all consensus methods here

load_consensus_map
make_consensus_tree
etc...
"""
from t2t.nlevel import RANK_ORDER
from numpy import zeros, where, logical_or, long


def taxa_score(master, reps):
    """Score taxa strings by contradictions observed in reps"""
    n_ranks = len(RANK_ORDER)
    master_ids = frozenset(master.keys())

    master_order = master.keys()
    master_rows = {k: idx for idx, k in enumerate(master_order)}
    master_cons = [master[k] for k in master_order]

    scores = zeros((len(master_ids), n_ranks), dtype=float)

    n_reps = 0
    for rep in reps:
        n_reps += 1

        for id_, con in rep.items():
            if id_ not in master_ids:
                raise KeyError("Unknown key %s in replicate" % id_)

            row = master_rows[id_]

            for rank, name in enumerate(con):
                if name == master_cons[row][rank]:
                    scores[row, rank] += 1

        # missing taxa are not considered contradictions
        missing_taxa = master_ids - frozenset(rep)
        for k in missing_taxa:
            row = master_rows[k]
            scores[row] += 1

    scores /= n_reps

    # slice and dice the scores
    return {k: scores[master_rows[k]] for k in master}


def merge_taxa_strings_and_scores(master, scores):
    """Merge taxa strings and their scores, return {id_:(taxa,score)}"""
    return {k: list(zip(v, scores[k])) for k, v in master.items()}


def taxa_score_hash(master, reps):
    """Score each taxonomy string based on contradictions observed in reps"""
    n_ranks = len(RANK_ORDER)

    master_order = master.keys()
    scores = zeros((len(master_order), n_ranks), dtype=float)
    master_hash = hash_cons(master, master_order, n_ranks)

    n_reps = 0
    for rep in reps:
        n_reps += 1
        rep_hash = hash_cons(rep, master_order, n_ranks)
        # where the taxons are equal to the master
        # or if the the taxon is not present in the replicate
        scores += where(logical_or(rep_hash == master_hash, rep_hash == 0),
                        1, 0)
    scores /= n_reps

    # slice and dice the scores
    return {k: scores[idx] for idx, k in enumerate(master_order)}


def hash_cons(cons, order, n_ranks):
    """Returns a numpy array of hash values for the cons

    NOTE: expects that cons are always specified even if the taxon name does
    not exist. In other words, the following are acceptable for missing fieids:
    [None, 'None', k__, p__, etc...]. It is _NOT_ okay to use the empty string
    at a field. The python hash method returns 0 on an empty string, but never
    otherwise and this method treats 0 specially.
    """
    hashes = zeros((len(order), n_ranks), dtype=long)

    for idx, id_ in enumerate(order):
        try:
            hashes[idx] = [hash(c) for c in cons[id_]]
        except KeyError:
            pass
        # defaults to zero if the consensus string isn't represented
    return hashes


def get_consensus_stats(consensus_map):
    """Returns consensus stats, expects rank prefix

    Returns a tuple of two dicts:

    - sequence counts per level, (classified, unclassified)
    - contains name counts per level
    """
    n_seqs = {}
    n_names = {}

    cons = consensus_map.values()
    total_cons = len(cons)

    rank_names = None
    for co in cons:
        if co is None:
            continue
        rank_names = [c[0] for c in co if c is not None]
        break
    for idx, rank in enumerate(rank_names):
        # collect all cons that are classified (ie more info than k__)
        klassed = [c[idx].lower()
                   for c
                   in cons if c[idx] and c[idx][0] == rank]

        n_classified = len(klassed)

        rank_names = set(klassed)

        n_seqs[rank] = (n_classified, total_cons - n_classified)
        n_names[rank] = rank_names

    return (n_seqs, n_names)


def pretty_print_consensus_stats(stats):
    seqs, names = stats
    print('\t'.join(['rank', 'num_classified', 'num_unclassified',
                     'num_names']))
    for k in RANK_ORDER:
        print('\t'.join(map(str, [k, seqs[k][0], seqs[k][1], len(names[k])])))



from skbio import TreeNode

import t2t.nlevel as nl
import t2t.validate as val


def fetch(tree):
    t = TreeNode.from_newick(open(tree))
    ranks = set(nl.RANK_ORDER)
    res = []
    error = True

    failed_nodes = []
    for n in t.non_tips():
        if n.name and n.name[0] not in ranks:
            failed_nodes.append(n)

    if failed_nodes:
        for n in failed_nodes:
            tip = list(n.tips())[0]
            tip_id = tip.name
            path = [a.name for a in tip.ancestors() if a.name is not None]
            res.append("Unknown rank: %s" % n.name)
            res.append("\tA tip ID from the clade: %s" % tip_id)
            res.append("\tCurrent lineage in tree: %s" % '; '.join(path[::-1]))

    else:
        res = nl.pull_consensus_strings(t)
        error = False

    return res, error


def validate(lines, limit, flat_errors, hierarchy_errors):
    res = []
    if flat_errors:
        flat = val.flat_errors(lines)
        for err_type in sorted(flat):
            ids = ','.join(flat[err_type][:10])

            if len(flat[err_type]) > limit:
                ellipse = '...'
            else:
                ellipse = ''

            res.append(err_type)
            res.append('\t%s%s' % (ids, ellipse))

    if hierarchy_errors:
        hier = val.hierarchy_errors(lines)
        if hier:
            res.append("Multiple parents")
        for err in hier:
            res.append("\t%s" % err['Taxon'])
            for parent in err['Parents']:
                res.append("\t\t%s, %s" % (parent, err['Parents'][parent]))

    return res, False



#!/usr/bin/env python

from unittest import TestCase, main
from t2t.util import reroot, unzip
from skbio import TreeNode

import sys

if sys.version_info[0] == 2:
    from StringIO import StringIO
else:
    from io import StringIO

__author__ = "Daniel McDonald"
__copyright__ = "Copyright 2011, The tax2tree project"
__credits__ = ["Daniel McDonald"]
__license__ = "BSD"
__version__ = "1.0"
__maintainer__ = "Daniel McDonald"
__email__ = "mcdonadt@colorado.edu"
__status__ = "Development"


class UtilTests(TestCase):

    def setUp(self):
        pass

    def test_reroot(self):
        """Should correctly reroot a tree"""
        t = TreeNode.read(StringIO(u"(((a,b)c,(d,e)f)g,(h,i)j);"))
        tips = ['a', 'b']
        for n in t.traverse():
            n.Length = 1.0

        # note, g is lost because it has a single descendent and gets pruned
        # off
        exp = ("((a:1.0,b:1.0)c:0.5,((d:1.0,e:1.0)f:1.0,(h:1.0,i:1.0)"
               "j:2.0):0.5);")
        exp = "((a,b)c,((d,e)f,(h,i)j):0.0);"
        obs = reroot(t, tips)

        self.assertEqual(str(obs).rstrip(), exp)

    def test_unzip(self):
        """unzip(items) should be the inverse of zip(*items)"""
        chars = [list('abcde'), list('ghijk')]
        numbers = [[1, 2, 3, 4, 5], [0, 0, 0, 0, 0]]
        strings = [["abcde", "fghij", "klmno"], ['xxxxx'] * 3]

        lists = [chars, numbers, strings]
        zipped = [zip(*i) for i in lists]
        unzipped = [unzip(i) for i in zipped]

        for u, l in zip(unzipped, lists):
            self.assertEqual(u, l)

if __name__ == '__main__':
    main()



#!/usr/bin/env python

"""nlevel tests

While many of the tests utilize similar trees and input data, the overlap
is not necessarily 100%. Many of these inputs are written with specific tests
in mind.
"""

__author__ = "Daniel McDonald"
__copyright__ = "Copyright 2011, The tax2tree project"
__credits__ = ["Daniel McDonald"]
__license__ = "BSD"
__version__ = "1.0"
__maintainer__ = "Daniel McDonald"
__email__ = "mcdonadt@colorado.edu"
__status__ = "Development"

from unittest import TestCase, main
from t2t.nlevel import (load_consensus_map, collect_names_at_ranks_counts,
                        load_tree, decorate_name_relative_freqs, decorate_name_counts,
                        set_ranksafe,
                        pick_names, has_badname, get_nearest_named_ancestor,
                        walk_consensus_tree, make_consensus_tree,
                        backfill_names_gap, commonname_promotion,
                        decorate_ntips, decorate_ntips_rank,
                        name_node_score_fold,
                        validate_all_paths, score_tree)

from skbio import TreeNode
import sys

if sys.version_info[0] == 2:
    from StringIO import StringIO
else:
    from io import StringIO

class NLevelTests(TestCase):

    def setUp(self):
        pass

    def test_score_tree(self):
        """Determine's the tree's fmeasure score"""
        # set RankNames and RankNameScores
        # if name in RankNames, check score, look at tips, etc
        t_str = StringIO(u"(((a,b),(c,d))e,(f,g),h)i;")
        t = TreeNode.read(t_str)
        t.RankNames = ['i', None, None, None]  # 1.0 * 6
        t.RankNameScores = [1.0, None, None, None]
        t.children[0].RankNames = [None, 'e', 'foo', None]  # 0.5 * 3, 0.6 * 3
        t.children[0].RankNameScores = [None, 0.5, 0.6, None]
        t.children[0].children[0].RankNames = [None] * 7
        t.children[0].children[1].RankNames = [None] * 7
        t.children[1].RankNames = [None] * 7
        t.children[1].RankNameScores = [None] * 7
        tips = list(t.tips())
        tips[0].Consensus = [None] * 7
        tips[1].Consensus = [1, 3, None, None]
        tips[2].Consensus = [2, 4, 5, None]
        tips[3].Consensus = [None, 1, None, None]
        tips[4].Consensus = [None, 1, None, None]
        tips[5].Consensus = [2, None, 3, None]
        tips[6].Consensus = [None, 4, None, None]
        decorate_ntips(t)
        exp = ((1.0 * 6) + (0.5 * 3) + (0.6 * 3)) / (6 + 3 + 3)
        obs = score_tree(t)
        self.assertEqual(obs, exp)

    def test_has_badname(self):
        """correctly determines if a string is bad"""
        data = "assadasd environmental sample asdasdd"
        self.assertTrue(has_badname(data))
        data = "asdasdsad dsasda dasd as"
        self.assertFalse(has_badname(data))

    def test_load_consensus_map(self):
        """correctly returns a consensus map"""
        data = ["foo\ta; b; c; d; e; f; g",
                 "bar\th; i; j; k; l; m; n",
                 "foobar\th; i; j; None; l; ; foo uncultured bar"]
        exp_noappend = {'foo': ['a', 'b', 'c', 'd', 'e', 'f', 'g'],
                        'bar': ['h', 'i', 'j', 'k', 'l', 'm', 'n'],
                        'foobar': ['h', 'i', 'j', None, 'l', None, None]}
        exp_append = {
            'foo': ['d__a', 'p__b', 'c__c', 'o__d', 'f__e', 'g__f', 's__g'],
            'bar': ['d__h', 'p__i', 'c__j', 'o__k', 'f__l', 'g__m', 's__n'],
            'foobar': ['d__h', 'p__i', 'c__j', 'o__', 'f__l', 'g__', 's__']}
        obs_noappend = load_consensus_map(data, False)
        obs_append = load_consensus_map(data, True)
        self.assertEqual(obs_noappend, exp_noappend)
        self.assertEqual(obs_append, exp_append)

    def test_collect_names_at_ranks_counts(self):
        """correctly returns total counts for names at ranks"""
        data = StringIO(u"((a,b)c,(d,(e,f)g)h,(i,j)k)l;")
        tipname_map = {'a': ['1', '2', '3', '4', '5', '6', '7'],
                       'b': ['1', '2', '3', None, '5', '6', '8'],
                       'd': ['1', '2', '3', 'a', '5', '6', '9'],
                       'e': ['1', '2', '3', None, '5', '6', '9'],
                       'i': ['1', '2', '3', 'a', '5', '6', '9'],
                       'j': ['1', '2', '3', '4', '5', '6', '9']}
        tree = load_tree(data, tipname_map)

        exp = {0: {'1': 6},
               1: {'2': 6},
               2: {'3': 6},
               3: {'4': 2, 'a': 2},
               4: {'5': 6},
               5: {'6': 6},
               6: {'7': 1, '8': 1, '9': 4}}

        obs = collect_names_at_ranks_counts(tree)
        self.assertEqual(obs, exp)

    def test_load_tree(self):
        """correctly loads and decorates tiplook info on a tree"""
        data = StringIO(u"((a,b)c,(d,(e,f)g)h,(i,j)k)l;")
        tipname_map = {'a': ['1', '2', '3', '4', '5', '6', '7'],
                       'b': ['1', '2', '3', '4', '5', '6', '8'],
                       'd': ['1', '2', '3', '4', '5', '6', '9'],
                       'e': ['1', '2', '3', '4', '5', '6', '10'],
                       'i': ['1', '2', '3', '4', '5', '6', '12'],
                       'j': ['1', '2', '3', '4', '5', '6', '13']}

        # exp in Name: (tipstart, tipstop, consensus)
        exp = {'a': (0, 0, ['1', '2', '3', '4', '5', '6', '7']),
               'b': (1, 1, ['1', '2', '3', '4', '5', '6', '8']),
               'c': (0, 1, [None] * 7),
               'd': (2, 2, ['1', '2', '3', '4', '5', '6', '9']),
               'e': (3, 3, ['1', '2', '3', '4', '5', '6', '10']),
               'f': (4, 4, [None] * 7),
               'g': (3, 4, [None] * 7),
               'h': (2, 4, [None] * 7),
               'i': (5, 5, ['1', '2', '3', '4', '5', '6', '12']),
               'j': (6, 6, ['1', '2', '3', '4', '5', '6', '13']),
               'k': (5, 6, [None] * 7),
               'l': (0, 6, [None] * 7)}

        obstree = load_tree(data, tipname_map)
        obs = {}
        for node in obstree.traverse(include_self=True):
            obs[node.name] = (node.TipStart, node.TipStop, node.Consensus)

        self.assertEqual(obs, exp)

    def test_decorate_name_relative_freqs(self):
        """correctly decorate relative frequency information on a tree"""
        data = StringIO(u"((a,b)c,(d,(e,f)g)h,(i,j)k)l;")
        tipname_map = {'a': ['1', '2', '3', '4', '5', '6', '7'],
                       'b': ['1', '2', '3', '4', '5', '6', '8'],
                       'd': ['1', '2', '3', '4', '5', '6', '8'],
                       'e': ['1', '2', '3', '4', 'a', '6', '7'],
                       'i': ['1', '2', '3', '4', 'a', None, '7'],
                       'j': ['1', '2', '3', '4', 'a', None, '8']}

        total_counts = {0: {'1': 10, 'foo': 5},
                        1: {'2': 6},
                        2: {'3': 12},
                        3: {'4': 6, 'bar': 5},
                        4: {'5': 6, 'a': 3},
                        5: {'6': 6},
                        6: {'7': 3, '8': 3}}

        tree = load_tree(data, tipname_map)
        decorate_ntips(tree)
        decorate_name_relative_freqs(tree, total_counts, 1)

        exp_root = {0: {'1': .6},
                    1: {'2': 1.0},
                    2: {'3': .5},
                    3: {'4': 1.0},
                    4: {'5': .5, 'a': 1.0},
                    5: {'6': 4.0 / 6},
                    6: {'7': 1.0, '8': 1.0}}

        self.assertEqual(tree.ConsensusRelFreq, exp_root)

    def test_decorate_name_counts(self):
        """correctly decorate relative frequency information on a tree"""
        data = StringIO(u"((a,b)c,(d,(e,f)g)h,(i,j)k)l;")
        tipname_map = {'a': ['1', '2', '3', '4', '5', '6', '7'],
                       'b': ['1', '2', '3', '4', '5', '6', '8'],
                       'd': ['1', '2', '3', '4', '5', '6', '8'],
                       'e': ['1', '2', '3', '4', 'a', '6', '7'],
                       'i': ['1', '2', '3', '4', 'a', None, '7'],
                       'j': ['1', '2', '3', '4', 'a', None, '8']}

        tree = load_tree(data, tipname_map)
        decorate_ntips(tree)
        decorate_name_counts(tree)

        exp_root = {0: {'1': 6},
                    1: {'2': 6},
                    2: {'3': 6},
                    3: {'4': 6},
                    4: {'5': 3, 'a': 3},
                    5: {'6': 4},
                    6: {'7': 3, '8': 3}}

        self.assertEqual(tree.TaxaCount, exp_root)

    def test_set_ranksafe(self):
        """correctly set ranksafe on tree"""
        data = StringIO(u"((a,b)c,(d,(e,f)g)h,(i,j)k)l;")
        tipname_map = {'a': ['1', '2', '3', '4', '5', '6', '7'],
                       'b': ['1', '2', '3', 'b', '5', '6', '8'],
                       'd': ['1', '2', '3', '4', '5', '6', '8'],
                       'e': ['1', '2', '3', 'b', 'a', '6', '7'],
                       'i': ['1', '2', '3', '4', 'a', '6', '7'],
                       'j': ['1', '2', '3', 'b', 'a', '6', '8']}

        total_counts = {0: {'1': 10, 'foo': 5},
                        1: {'2': 20},
                        2: {'3': 10},
                        3: {'4': 4, 'b': 5},
                        4: {'5': 7, 'a': 5},
                        5: {'6': 6},
                        6: {'7': 3, '8': 3}}

        tree = load_tree(data, tipname_map)
        decorate_ntips(tree)
        decorate_name_relative_freqs(tree, total_counts, 1)
        set_ranksafe(tree)

        exp_root = [True, False, True, False, True, True, False]
        self.assertEqual(tree.RankSafe, exp_root)

    def test_name_node_score_fold(self):
        """hate taxonomy"""
        data = StringIO(u"((a,b)c,(d,(e,f)g)h,(i,j)k)l;")
        tipname_map = {'a': ['1', '2', '3', '4', '5', '6', '8'],
                       'b': ['1', '2', '3', '4', '5', '6', '8'],
                       'd': ['1', '2', '3', 'f', 'e', 'c', '9'],
                       'e': ['1', '2', '3', 'f', 'e', 'c', '9'],
                       'i': ['1', '2', '3', 'g', 'a', 'h', '11'],
                       'j': ['1', '2', '3', 'g', 'a', 'h', '12']}

        total_counts = {0: {'1': 10, 'foo': 5},
                        1: {'2': 10},
                        2: {'3': 10},
                        3: {'4': 3, 'f': 5, 'g': 5},
                        4: {'5': 7, 'a': 5, 'e': 4},
                        5: {'6': 3, 'c': 3, 'd': 2, 'h': 3},
                        6: {'8': 3, '9': 2, '10': 2, '11': 2, '12': 2}}

        tree = load_tree(data, tipname_map)
        decorate_ntips(tree)
        decorate_name_relative_freqs(tree, total_counts, 1)
        set_ranksafe(tree)
        pick_names(tree)
        name_node_score_fold(tree)
        exp_root = ['1', '2', '3', None, None, None, None]
        expc0 = [None, None, None, '4', None, None, None]
        expc1 = [None, None, None, None, 'e', 'c', '9']
        expc2 = [None, None, None, None, None, 'h', None]
        expc1c1 = [None] * 7

        self.assertEqual(tree.RankNames, exp_root)
        self.assertEqual(tree.children[0].RankNames, expc0)
        self.assertEqual(tree.children[1].RankNames, expc1)
        self.assertEqual(tree.children[2].RankNames, expc2)
        self.assertEqual(tree.children[1].children[1].RankNames, expc1c1)

    def test_validate_all_paths(self):
        """complains correctly about badpaths"""
        data = StringIO(u"(((((1,2)s__,(3,4)s__)g__)p__),((5,6)f__)f__,((7,8)c__)o__);")
        t = load_tree(data, {})
        exp = [n for n in t.tips() if n.name in ['5', '6', '7', '8']]
        obs = validate_all_paths(t)
        self.assertEqual(obs, exp)

    def test_best_name_freqs_for_nodes(self):
        """correctly gets the frequencies per name per node"""
        data = StringIO(u"((a,b)c,(d,(e,f)g)h,(i,j)k)l;")
        tipname_map = {'a': ['1', '2', '3', '4', '5', '6', '7'],
                       'b': ['1', '2', '3', 'b', '5', '6', '8'],
                       'd': ['1', '2', '3', '4', '5', '6', '7'],
                       'e': ['1', '2', '3', 'b', 'a', 'foo', '7'],
                       'i': ['1', '2', '3', '4', 'a', 'foo', '8'],
                       'j': ['1', '2', '3', 'b', 'a', 'foo', '8']}

        total_counts = {0: {'1': 10, 'foo': 5},
                        1: {'2': 10},
                        2: {'3': 10},
                        3: {'4': 4, 'b': 5},
                        4: {'5': 7, 'a': 5},
                        5: {'6': 3, 'foo': 3},
                        6: {'7': 3, '8': 4}}

        tree = load_tree(data, tipname_map)
        decorate_ntips(tree)
        decorate_name_relative_freqs(tree, total_counts, 1)
        set_ranksafe(tree)
        pick_names(tree)

        # result = best_name_freqs_for_nodes(tree)
        cnode = tree.children[0]
        hnode = tree.children[1]
        knode = tree.children[2]
        # exp = {0:{'1':[(0.6,tree)]}
        #        1:{'2':[(0.6,tree)]},
        #        2:{'3':[(0.6,tree)]},
        #        3:{'4':[(0.75,tree)],'b':[(0.6,tree)]},
        #        4:{'a':[(0.6, tree),[(0.5,knode)]

    def test_pick_names(self):
        """correctly pick names to retain on a tree"""
        data = StringIO(u"((a,b)c,(d,(e,f)g)h,(i,j)k)l;")
        tipname_map = {'a': ['1', '2', '3', '4', '5', '6', '7'],
                       'b': ['1', '2', '3', 'b', '5', '6', '8'],
                       'd': ['1', '2', '3', '4', '5', '6', '7'],
                       'e': ['1', '2', '3', 'b', 'a', 'foo', '7'],
                       'i': ['1', '2', '3', '4', 'a', 'foo', '8'],
                       'j': ['1', '2', '3', 'b', 'a', 'foo', '8']}

        total_counts = {0: {'1': 10, 'foo': 5},
                        1: {'2': 10},
                        2: {'3': 10},
                        3: {'4': 4, 'b': 5},
                        4: {'5': 7, 'a': 5},
                        5: {'6': 3, 'foo': 3},
                        6: {'7': 3, '8': 4}}

        tree = load_tree(data, tipname_map)
        decorate_ntips(tree)
        decorate_name_relative_freqs(tree, total_counts, 1)
        set_ranksafe(tree)
        pick_names(tree)
        exp_root = ['1', '2', '3', None, None, None, None]
        self.assertEqual(tree.RankNames, exp_root)

        expc0 = [None, None, None, None, None, '6', None]
        expc1 = [None, None, None, None, None, None, '7']
        expc2 = [None, None, None, None, None, 'foo', '8']
        expc1c1 = [None] * 7

        self.assertEqual(tree.children[0].RankNames, expc0)
        self.assertEqual(tree.children[1].RankNames, expc1)
        self.assertEqual(tree.children[2].RankNames, expc2)
        self.assertEqual(tree.children[1].children[1].RankNames, expc1c1)

    def test_walk_consensus_tree(self):
        """correctly walk consensus tree"""
        data = [['a', 'b', 'c', 'd', 'e', 'f', 'g'],
                 ['a', 'b', 'c', None, None, 'x', 'y'],
                 ['h', 'i', 'j', 'k', 'l', 'm', 'n'],
                 ['h', 'i', 'j', 'k', 'l', 'm', 'q'],
                 ['h', 'i', 'j', 'k', 'l', 'm', 'n']]
        _, lookup = make_consensus_tree(data, check_for_rank=False)
        exp1 = ['n', 'm', 'l']
        exp2 = ['a']
        exp3 = ['x', 'f__', 'o__', 'c']
        obs1 = walk_consensus_tree(lookup, 'n', 3, reverse=False)
        obs2 = walk_consensus_tree(lookup, 'a', 3, reverse=False)
        obs3 = walk_consensus_tree(lookup, 'x', 4, reverse=False)
        self.assertEqual(obs1, exp1)
        self.assertEqual(obs2, exp2)
        self.assertEqual(obs3, exp3)

    def test_make_consensus_tree(self):
        """correctly gets parent consensus counts"""
        data = [['a', 'b', 'c', 'd', 'e', 'f', 'g'],
                 ['a', 'b', 'c', None, None, 'x', 'y'],
                 ['h', 'i', 'j', 'k', 'l', 'm', 'n'],
                 ['h', 'i', 'j', 'k', 'l', 'm', 'q'],
                 ['h', 'i', 'j', 'k', 'l', 'm', 'n']]
        exp_str = "(((((((g)f)e)d,(((y)x)))c)b)a,((((((n,q)m)l)k)j)i)h);"

        obs_root, lookup = make_consensus_tree(data, check_for_rank=False)

        self.assertEqual(str(obs_root).strip(), exp_str)
        self.assertNotIn(None, lookup)

    def test_make_consensus_tree_withtips(self):
        """correctly constructs the taxonomy tree with tip info"""
        data = [['a', 'b', 'c', 'd', 'e', 'f', 'g'],
                 ['a', 'b', 'c', None, None, 'x', 'y'],
                 ['h', 'i', 'j', 'k', 'l', 'm', 'n'],
                 ['h', 'i', 'j', 'k', 'l', 'm', 'q'],
                 ['h', 'i', 'j', 'k', 'l', 'm', 'n']]
        input_ids = ['1', '2', '3', '4', '5']
        exp_str = ("((((((((1)g)f)e)d,((((2)y)x)))c)b)a,(((((((3,5)n,"
                   "(4)q)m)l)k)j)i)h);")

        obs_root, lookup = make_consensus_tree(
            data, check_for_rank=False, tips=input_ids)

        self.assertEqual(str(obs_root).strip(), exp_str)
        self.assertNotIn(None, lookup)

    def test_decorate_ntips(self):
        """correctly decorate the tree with the NumTips param"""
        data = StringIO(u"(((a,b)c,(d,e,f)g)h,(i,j)k)l;")
        tree = TreeNode.read(data)
        tips = dict([(tip.name, tip) for tip in tree.tips()])
        tips['a'].Consensus = [1, 2, 3, 4, 5, 6, 7]
        tips['b'].Consensus = [None, None, None, 5, None, None, None]
        tips['d'].Consensus = [1, 2, 3, 4, 5, 6, 8]
        tips['e'].Consensus = [None, None, None, None, None, None, None]
        tips['f'].Consensus = [1, 2, 3, 4, 5, 6, 8]
        tips['i'].Consensus = [1, 2, 3, 4, 5, 6, 8]
        tips['j'].Consensus = [1, 2, 3, 4, 5, 6, 8]
        decorate_ntips(tree)
        self.assertEqual(tree.NumTips, 6)
        self.assertEqual(tree.children[0].NumTips, 4)
        self.assertEqual(tree.children[1].NumTips, 2)
        self.assertEqual(tree.children[0].children[0].NumTips, 2)
        self.assertEqual(tree.children[0].children[1].NumTips, 2)

    def test_decorate_ntips_rank(self):
        """correctly decorate the tree with the NumTipsRank param"""
        data = StringIO(u"(((a,b)c,(d,e,f)g)h,(i,j)k)l;")
        tree = TreeNode.read(data)
        tips = dict([(tip.name, tip) for tip in tree.tips()])
        tips['a'].Consensus = [1, 2, 3, 4, 5, 6, 7]
        tips['b'].Consensus = [None, None, None, 5, None, None, None]
        tips['d'].Consensus = [1, 2, 3, 4, 5, 6, 8]
        tips['e'].Consensus = [None, None, None, None, None, None, None]
        tips['f'].Consensus = [1, 2, 3, 4, 5, 6, 8]
        tips['i'].Consensus = [1, 2, 3, 4, 5, 6, 8]
        tips['j'].Consensus = [1, 2, 3, 4, 5, 6, 8]
        decorate_ntips_rank(tree)

        self.assertEqual(tree.NumTipsRank[0], 5)
        self.assertEqual(tree.NumTipsRank[1], 5)
        self.assertEqual(tree.NumTipsRank[2], 5)
        self.assertEqual(tree.NumTipsRank[3], 6)
        self.assertEqual(tree.NumTipsRank[4], 5)
        self.assertEqual(tree.NumTipsRank[5], 5)
        self.assertEqual(tree.NumTipsRank[6], 5)

    def test_get_nearest_named_ancestor(self):
        """correctly get the nearest named ancestor"""
        t = TreeNode.read(StringIO(u"(((s1,s2)g1,s3))root;"))
        t2 = TreeNode.read(StringIO(u"(((s1,s2)g1,s3));"))
        exp_t = t
        exp_t2 = None
        obs_t = get_nearest_named_ancestor(t.find('s3'))
        obs_t2 = get_nearest_named_ancestor(t2.find('s3'))
        self.assertEqual(obs_t, exp_t)
        self.assertEqual(obs_t2, exp_t2)

    def test_backfill_names_gap(self):
        """correctly backfill names"""
        consensus_tree = TreeNode.read(StringIO(u"(((s1,s2)g1,(s3,s4)g2,(s5,s6)g3)f1)o1;"))
        rank_lookup = {'s': 6, 'g': 5, 'f': 4, 'o': 3, 'c': 2, 'p': 1, 'k': 0}
        for n in consensus_tree.traverse(include_self=True):
            n.Rank = rank_lookup[n.name[0]]
        data = StringIO(u"((((1)s1,(2)s2),((3)s3,(4)s5)))o1;")
        lookup = dict([(n.name, n)
                      for n in consensus_tree.traverse(include_self=True)])
        # exp = "((((1)s1,(2)s2)g1,((3)'g2; s3',(4)'g3; s5')))'o1; f1'"
        t = TreeNode.read(data)
        t.Rank = 3
        t.children[0].Rank = None
        t.children[0].children[0].Rank = None
        t.children[0].children[1].Rank = None
        t.children[0].children[0].children[0].Rank = 6
        t.children[0].children[0].children[1].Rank = 6
        t.children[0].children[1].children[0].Rank = 6
        t.children[0].children[1].children[1].Rank = 6

        backfill_names_gap(t, lookup)

        self.assertEqual(t.BackFillNames, ['o1'])
        self.assertEqual(t.children[0].BackFillNames, [])
        self.assertEqual(t.children[0].children[0].BackFillNames, [])
        self.assertEqual(t.children[0].children[1].BackFillNames, [])
        self.assertEqual(t.children[0].children[0]
                         .children[0].BackFillNames, ['f1', 'g1', 's1'])
        self.assertEqual(t.children[0].children[0]
                         .children[1].BackFillNames, ['f1', 'g1', 's2'])
        self.assertEqual(t.children[0].children[1]
                         .children[0].BackFillNames, ['f1', 'g2', 's3'])
        self.assertEqual(t.children[0].children[1]
                         .children[1].BackFillNames, ['f1', 'g3', 's5'])

    def test_backfill_names_dangling(self):
        """correctly fill in dangling missing ranks"""
        consensus_tree = TreeNode.read(StringIO(u"(((s1,s2)g1,(s3,s4)g2,(s5,s6)g3)f1)o1;"))
        data = "((((1),(2)),((3),(4))))'o1; f1';"
        lookup = dict([(n.name, n)
                      for n in consensus_tree.traverse(include_self=True)])
        # exp = "((((1),(2)),((3),(4))))'o1; f1';"

    def test_commonname_promotion(self):
        """correctly promote names if possible"""
        consensus_tree = TreeNode.read(StringIO(u"(((s1,s2)g1,(s3,s4)g2,(s5,s6)g3)f1)o1;"))
        rank_lookup = {'s': 6, 'g': 5, 'f': 4, 'o': 3, 'c': 2, 'p': 1, 'k': 0}
        for n in consensus_tree.traverse(include_self=True):
            n.Rank = rank_lookup[n.name[0]]
        data = StringIO(u"((((1)s1,(2)s2),((3)s3,(4)s5)))o1;")
        lookup = dict([(n.name, n)
                      for n in consensus_tree.traverse(include_self=True)])
        exp = "((((1)s1,(2)s2)g1,((3)'g2; s3',(4)'g3; s5')))'o1; f1';"
        t = TreeNode.read(data)
        t.Rank = 3
        t.children[0].Rank = None
        t.children[0].children[0].Rank = None
        t.children[0].children[1].Rank = None
        t.children[0].children[0].children[0].Rank = 6
        t.children[0].children[0].children[1].Rank = 6
        t.children[0].children[1].children[0].Rank = 6
        t.children[0].children[1].children[1].Rank = 6
        backfill_names_gap(t, lookup)
        commonname_promotion(t)

        self.assertEqual(str(t).rstrip(), exp)

if __name__ == '__main__':
    main()



#!/usr/bin/env python
from __future__ import division
from setuptools import setup
import os

__author__ = "Daniel McDonald"
__copyright__ = "Copyright 2011, The tax2tree project"
__credits__ = ["Daniel McDonald"]
__license__ = "BSD"
__version__ = "1.0"
__maintainer__ = "Daniel McDonald"
__email__ = "mcdonadt@colorado.edu"
__status__ = "Development"

classes = """
    Development Status :: 4 - Beta
    License :: OSI Approved :: BSD License
    Topic :: Scientific/Engineering
    Topic :: Scientific/Engineering :: Bio-Informatics
    Programming Language :: Python
    Programming Language :: Python :: 3.6
    Operating System :: Unix
    Operating System :: POSIX
    Operating System :: MacOS :: MacOS X
"""
classifiers = [s.strip() for s in classes.split('\n') if s]

long_description = """The tax2tree project"""

# If readthedocs.org is building the project, we're not able to build the
# required numpy/scipy versions on their machines (nor do we want to, as that
# would take a long time). To build the docs, we don't need the latest versions
# of these dependencies anyways, so we use whatever is in their system's
# site-packages to make scikit-bio importable. See doc/rtd-requirements.txt for
# dependencies that RTD must install in order to build our docs.
#
# Code to check whether RTD is building our project is taken from
# http://read-the-docs.readthedocs.org/en/latest/faq.html
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
if on_rtd:
    install_requires = []
else:
    install_requires = ['numpy >= 1.7', 'future>=0.13.1', 'scikit-bio',
                        'Click']

setup(name='tax2tree',
      version=__version__,
      description='Taxonomy to tree decoration tools',
      author=__maintainer__,
      author_email=__email__,
      maintainer=__maintainer__,
      maintainer_email=__email__,
      url='https://github.com/biocore/tax2tree',
      packages=['t2t'],
      scripts=['scripts/t2t'],
      install_requires=install_requires,
      extras_require={'test': ['nose >= 0.10.1', 'pep8'],
                      'doc': ['Sphinx >= 1.2.2']},
      long_description=long_description)



#!/usr/bin/env python

"""consistency tests

While many of the tests utilize similar trees and input data, the overlap
is not necessarily 100%. Many of these inputs are written with specific tests
in mind.
"""

__author__ = "Donovan Park"
__copyright__ = "Copyright 2014, The tax2tree project"
__credits__ = ["Donovan Park"]
__license__ = "BSD"
__version__ = "1.0"
__maintainer__ = "Donovan Park"
__email__ = "donovan.parks@gmail.com"
__status__ = "Development"

from unittest import TestCase, main

import t2t.nlevel as nl
from t2t.consistency import Consistency

from io import StringIO

class ConsistencyTests(TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        nl.set_rank_order(['d', 'p', 'c', 'o', 'f', 'g', 's'])

    def test_consistency_missing(self):
        """Test consistency of taxa in tree with missing taxa"""

        seed_con = 'f__Lachnospiraceae; g__Bacteroides; s__'
        nl.determine_rank_order(seed_con)
        tipname_map = {'a': ['f__Lachnospiraceae', 'g__Bacteroides', None],
                       'c': ['f__Lachnospiraceae', 'g__Bacteroides', 's__Bacteroides pectinophilus'],
                       'b': ['f__Lachnospiraceae', 'g__Bacteroides', None], 'e': [None, None, None],
                       'd': ['f__Lachnospiraceae', 'g__Bacteroides', 's__Bacteroides pectinophilus'],
                       'g': [None, None, None], 'f': ['f__Lachnospiraceae', 'g__Lachnospira', None],
                       'h': ['f__Lachnospiraceae', 'g__Lachnospira', 's__Bacteroides pectinophilus']}
        tree = nl.load_tree(StringIO(u'(((a,b),(c,d)),((e,f),(g,h)));'), tipname_map)

        counts = nl.collect_names_at_ranks_counts(tree)
        nl.decorate_ntips_rank(tree)
        nl.decorate_name_counts(tree)

        # determine taxonomic consistency of rooted tree
        #expected_consistency_index
        c = Consistency(counts, len(nl.RANK_ORDER))
        consistency_index = c.calculate(tree, rooted=True)

        self.assertAlmostEqual(consistency_index[0]['f__Lachnospiraceae'], 1.0)
        self.assertAlmostEqual(consistency_index[1]['g__Bacteroides'], 1.0)
        self.assertAlmostEqual(consistency_index[1]['g__Lachnospira'], 1.0)
        self.assertAlmostEqual(consistency_index[2]['s__Bacteroides pectinophilus'], 1.0)

        #determine consistency of unrooted tree
        consistency_index = c.calculate(tree, rooted=False)

        self.assertAlmostEqual(consistency_index[0]['f__Lachnospiraceae'], 1.0)
        self.assertAlmostEqual(consistency_index[1]['g__Bacteroides'], 1.0)
        self.assertAlmostEqual(consistency_index[1]['g__Lachnospira'], 1.0)
        self.assertAlmostEqual(consistency_index[2]['s__Bacteroides pectinophilus'], 1.0)

    def test_consistency_unrooted(self):
        """Test consistency of taxa with a taxa that is only monophyletic in unrooted tree"""

        seed_con = 'f__Lachnospiraceae; g__Bacteroides; s__'
        nl.determine_rank_order(seed_con)
        tipname_map = {'a': ['f__Lachnospiraceae', 'g__Bacteroides', 's__Bacteroides pectinophilus'],
                       'b': ['f__Lachnospiraceae', 'g__Bacteroides', 's__Bacteroides pectinophilus'],
                       'c': ['f__Lachnospiraceae', 'g__Bacteroides', 's__Bacteroides pectinophilus'],
                       'd': ['f__Lachnospiraceae', 'g__Bacteroides', 's__Bacteroides acidifaciens'],
                       'e': ['f__Lachnospiraceae', 'g__Bacteroides', 's__Bacteroides acidifaciens']}

        tree = nl.load_tree(StringIO(u'((a,b),(c,(d,e)));'), tipname_map)

        counts = nl.collect_names_at_ranks_counts(tree)
        nl.decorate_ntips_rank(tree)
        nl.decorate_name_counts(tree)

        # determine taxonomic consistency of rooted tree
        #expected_consistency_index
        c = Consistency(counts, len(nl.RANK_ORDER))
        consistency_index = c.calculate(tree, rooted=True)

        self.assertAlmostEqual(consistency_index[0]['f__Lachnospiraceae'], 1.0)
        self.assertAlmostEqual(consistency_index[1]['g__Bacteroides'], 1.0)
        self.assertAlmostEqual(consistency_index[2]['s__Bacteroides pectinophilus'], 0.66666666)
        self.assertAlmostEqual(consistency_index[2]['s__Bacteroides acidifaciens'], 1.0)

        #determine consistency of unrooted tree
        consistency_index = c.calculate(tree, rooted=False)

        self.assertAlmostEqual(consistency_index[0]['f__Lachnospiraceae'], 1.0)
        self.assertAlmostEqual(consistency_index[1]['g__Bacteroides'], 1.0)
        self.assertAlmostEqual(consistency_index[2]['s__Bacteroides pectinophilus'], 1.0)
        self.assertAlmostEqual(consistency_index[2]['s__Bacteroides acidifaciens'], 1.0)

if __name__ == '__main__':
    main()
